//
// Copyright (C) 1999-2002 Toshikaz Hirabayashi
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to
// deal in the Software without restriction, including without limitation the
// rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
// sell copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
// TOSHIKAZ HIRABAYASHI BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
// WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF
// OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.
//
// Except as contained in this notice, the name of Toshikaz Hirabayashi shall
// not be used in advertising or otherwise to promote the sale, use or other
// dealings in this Software without prior written authorization from
// Toshikaz Hirabayashi.

#include <WScom.h>
#include <WSCclassInformation.h>
#include <WSCbaseList.h>
#include <WSCbase.h>
#include <WSClistData.h>

class WSCbaseArray{
 public:
  WSCbaseArray(char*);
  ~WSCbaseArray();
  long        _hash_value;
  char*       _cname;
  WSClistData _base_list;
  long     getNum();
  void     addItem(WSCbase*);
  void     delItem(WSCbase* );
  WSCbase* getItem(char* );
};
static WSCbaseList* _the_app_base_list = NULL;
WSCbaseList::WSCbaseList(){
  _search_handler = NULL;
  _base_list = new WSClistData(128);
  _need_update = new WSClistData(1024);
  _need_init = new WSClistData(1024);
}
WSCbaseList::~WSCbaseList(){
  long i;
  long num = _base_list->getNum();
  for(i=0; i < num; i++){
    WSCbaseArray* base_array = (WSCbaseArray*)(*_base_list)[i];
    delete base_array;
  }
  _base_list->clear();
  delete _base_list;
  delete _need_update;
  delete _need_init;
}
WSClistData* WSCbaseList::getInstanceList(char* cname){
  long i;
  long hvalue = WSGFgetHashValue(cname);
  long num = _base_list->getNum();
  for(i=0; i<num; i++){
    WSCbaseArray* base_array = (WSCbaseArray*)(*_base_list)[i];
    if (base_array->_hash_value == hvalue && !strcmp(base_array->_cname,cname)){
      return &base_array->_base_list;
    }
  }
  return NULL;
}
WSCbaseList* WSGIappObjectList(){
  if (_the_app_base_list == NULL){
    _the_app_base_list = new WSCbaseList();
  }
  return _the_app_base_list;
}
static WSCbase* _user_data_ = NULL;
void WSCbaseList::setUserData(char* name,void* data){
  if (_user_data_ == NULL){
    _user_data_ = new WSCbase(NULL,"_user_data_");
	_user_data_->initialize();
  }
  _user_data_->setUserData(name,data);
}
void* WSCbaseList::getUserData(char* name){
  if (_user_data_ != NULL){
    return _user_data_->getUserData(name);
  }
  return NULL;
}
WSClistData* WSCbaseList::getClassList(){
  return _base_list;
}
void WSCbaseList::addClient(WSCbase* client,void* base_array){
  //ok,it was initialized list in classinfo 
  if (base_array != NULL){
    WSCbaseArray* barray = (WSCbaseArray*)base_array;
    barray->addItem(client);
  }else{
    WSCclassInformation* inf = client->getClassInformation();
    WSCbaseArray* new_base_list = new WSCbaseArray(client->getClassName());
    new_base_list->addItem(client);
    _base_list->add( (void*) new_base_list );
    inf->setClassInstanceList((void*)new_base_list);
  }
}
void WSCbaseList::delClient(WSCbase* client,void* base_array){
  //delete from update list
  long i;
  long num = _need_update->getNum();
  for(i=0; i <num; i++){
    WSCbase* base = (WSCbase*)(*_need_update)[i];
    if (base == client){
      _need_update->delPos(i);
      num = _need_update->getNum();
      i--;
    }
  }
  //delete from initialize list
  num = _need_init->getNum();
  for(i=0; i <num; i++){
    WSCbase* base = (WSCbase*)(*_need_init)[i];
    if (base == client){
      _need_init->delPos(i);
      num = _need_init->getNum();
      i--;
    }
  }
  //delete from instance list
  WSCbaseArray* barray = (WSCbaseArray*)base_array;
  if (barray != NULL){
    barray->delItem(client);
  }else{
    WSCclassInformation* inf = client->getClassInformation();
    WSCbaseArray* barray2 = (WSCbaseArray*)inf->getClassInstanceList();
    if (barray2 != NULL){
      barray2->delItem(client);
    }
  }
}
WSCbase* WSCbaseList::searchInstance(char* cname,char* iname){
  if (_search_handler == NULL){
    return getInstance(cname,iname);
  }
  WSCbase* (*hd)(char*,char*) = (WSCbase* (*)(char*,char*))_search_handler;  
  return hd(cname,iname);
}
void WSCbaseList::setSearchInstanceHandler(void* hd){
  _search_handler = hd;
}

WSCbool WSCbaseList::existInstance(WSCbase* ptr){
  long num = _base_list->getNum();
  long i;
  for(i=0; i<num; i++){
    WSCbaseArray* base_array = (WSCbaseArray*)(*_base_list)[i];
    long num2 = base_array->getNum();
    long j;
    for(j=0; j < num2;j++){
      WSCbase* base = (WSCbase*)base_array->_base_list[j];
      if (base == ptr){
        return True;
      }
    }
  }
  return False;
}

WSCbase* WSCbaseList::getInstance(char* cname,char* iname){
  long hvalue = WSGFgetHashValue(cname);
  int i;
  int num = _base_list->getNum();
  if (!strcmp(cname,"WSCbase")){
    for(i = 0; i < num; i++){
      WSCbaseArray* base_array = (WSCbaseArray*)(*_base_list)[i];
      WSCbase* base =  base_array->getItem(iname);
      if (base != NULL){
        return base;
      }
    }
  }else{
    for(i = 0; i < num; i++){
      WSCbaseArray* base_array = (WSCbaseArray*)(*_base_list)[i];
      if (base_array->_hash_value == hvalue &&
          !strcmp(base_array->_cname,cname)){
        return base_array->getItem(iname);
      }
    }
  }
  return NULL;
}

void WSCbaseList::execUpdate(){
  long i;
  long num = _need_update->getNum();
  if (num > 0){
    WSCbase** baselist = new WSCbase*[num];
    memcpy(baselist,_need_update->getBuf(),sizeof(WSCbase*)*num);
    _need_update->clear();
    for(i=0; i <num; i++){
      if (baselist[i]->isNeedUpdate() != False){
        baselist[i]->update();
      }
    }
    delete baselist;
  }
}
void WSCbaseList::addNeedUpdate(WSCbase* client){
  _need_update->add( (void*)client );
}
void WSCbaseList::execInitialize(){
  long i;
  long num = _need_init->getNum();
  if (num > 0){
    WSCbase** baselist = new WSCbase*[num];
    memcpy(baselist,_need_init->getBuf(),sizeof(WSCbase*)*num);
    _need_init->clear();
    for(i=0; i <num; i++){
      baselist[i]->execProcedure(WSEV_INITIALIZE);
    }
    delete baselist;
  }
}
void WSCbaseList::addNeedInitialize(WSCbase* client){
  _need_init->add( (void*)client );
}
long WSCbaseList::countObject(char* cname){
  int i;
  long hvalue = WSGFgetHashValue(cname);
  int num = _base_list->getNum();
  if (!strcmp(cname,"WSCbase")){
    return num;
  }
  for(i = 0; i < num; i++){
    WSCbaseArray* base_array = (WSCbaseArray*)(*_base_list)[i];
    if (base_array->_hash_value == hvalue &&
        !strcmp(base_array->_cname,cname)){
      long ret = base_array->getNum();
      return ret;
    }
  }
  return 0;
}
WSCbaseArray::WSCbaseArray(char* cname){
  _cname = WSGFstrdup(cname);
  _hash_value = WSGFgetHashValue(cname);
}
WSCbaseArray::~WSCbaseArray(){
  _base_list.clear();
}
long WSCbaseArray::getNum(){
  return _base_list.getNum();
}
void WSCbaseArray::addItem(WSCbase* item){
  _base_list.add( (void*) item );
}
void WSCbaseArray::delItem(WSCbase* item){
  _base_list.del( (void*) item );
}
WSCbase* WSCbaseArray::getItem(char* item){
  long i;
  long hvalue = WSGFgetHashValue(item); 
  long num = _base_list.getNum();
  for(i=0; i < num; i++){
    WSCbase* base = (WSCbase*)_base_list[i];
    if ( hvalue == base->getInstanceNameHashValue() &&
      !strcmp(item,base->getInstanceName()) ){
      return base;
    }
  }
  return NULL;
}
void WSCbaseList::addEvent(WSCbase* client,long ev){
  _event_list.add((void*)ev);
  _event_client_list.add((void*)client);
}
void WSCbaseList::delEvent(WSCbase* client,long ev){
  long i;
  long num = _event_list.getNum(); 
  for(i=0; i<num; i++){
    WSCbase* cl = (WSCbase*)_event_client_list[i];
    if (cl == client){
      long e = (long)_event_list[i];
      if (e == ev){
        _event_list.delPos(i);
        _event_client_list.delPos(i);
        break;
      }
    }
  }
  _event_client_list.add((void*)client);
  _event_list.add((void*)ev);
}
void WSCbaseList::execEvent(long ev,void* data){
  long i;
  long num = _event_list.getNum(); 
  for(i=0; i<num; i++){
    long val = (long)_event_list[i];
    if (val == ev){
      WSCbase* client = (WSCbase*)_event_client_list[i];
      long tev = (long)_event_list[i];
      if (tev == ev){
        client->execEventProc(ev,data);
      }
    }
  }
}
void WSCbaseList::execEvent(long ev){
  long i;
  long num = _event_list.getNum(); 
  for(i=0; i<num; i++){
    long val = (long)_event_list[i];
    if (val == ev){
      WSCbase* client = (WSCbase*)_event_client_list[i];
      long tev = (long)_event_list[i];
      if (tev == ev){
        client->execProcedure(ev);
      }
    }
  }
}

void WSCbaseList::execEvent(char* pname){
  long i;
  long num = _event_list.getNum(); 
  for(i=0; i<num; i++){
    long val = (long)_event_list[i];
    WSCbase* client = (WSCbase*)_event_client_list[i];
    client->execProcedure(pname);
  }
}

