//
// Copyright (C) 1999-2002 Toshikaz Hirabayashi
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to
// deal in the Software without restriction, including without limitation the
// rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
// sell copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
// TOSHIKAZ HIRABAYASHI BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
// WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF
// OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.
//
// Except as contained in this notice, the name of Toshikaz Hirabayashi shall
// not be used in advertising or otherwise to promote the sale, use or other
// dealings in this Software without prior written authorization from
// Toshikaz Hirabayashi.

#include <WScom.h>
#include <WSCconductor.h>
#include <WSDserialize.h>
#include <WSCbase.h>

static WSCconductor* _the_app_conductor = NULL;
WSCconductor* WSGIconductor(){
  if (_the_app_conductor == NULL){
    _the_app_conductor = new WSCconductor();
  }
  return _the_app_conductor;
}
static WSCbinType* _the_app_bin_type = NULL;
WSCbinType* WSGIbinType(){
  if (_the_app_bin_type == NULL){
    _the_app_bin_type = new WSCbinType();
  }
  return _the_app_bin_type;
}

WSCconductor::WSCconductor(){
  _serialize_path.setString(".",WS_EN_ISO8859_1);
}
WSCconductor::~WSCconductor(){}

long WSCconductor::addSerialize(WSDserialize* serialize){
  _serialize_list.add(serialize);
  return WS_NO_ERR;
}
long WSCconductor::add(WSCtypeServer* item){
  _type_server_list.add(item);
  return WS_NO_ERR;
}
long WSCconductor::del(WSCtypeServer* item){
  _type_server_list.del(item);
  return WS_NO_ERR;
}
WSCtypeServer* WSCconductor::searchTypeServer(char* type){
  long i;
  long num = _type_server_list.getNum();
  for(i=0; i<num; i++){
    WSCtypeServer* drcl = (WSCtypeServer*)_type_server_list[i];
    if (drcl->isHit(type) != False){
      return drcl;
    }
  }
  return NULL;
}
WSDserialize* WSCconductor::getSerialize(char* serialize_type){
  long i;
  long num = _serialize_list.getNum();
  WSDserialize* same_serialize = NULL;
  for(i=0; i< num; i++){
    WSDserialize* serialize = (WSDserialize*)_serialize_list[i];
    if (serialize->type(serialize_type) != False ){
      same_serialize = serialize;
      if ( serialize->underTransaction() == False){
        return serialize;
      }
    }
  }
  if (same_serialize != NULL){
    WSDserialize* serialize = same_serialize->createSerialize();
    return serialize;
  }
WSMFtrace("WSCconductor::getSerialize()  ERROR. default type not found.\n");
  return NULL;
}
char* WSCconductor::getSerializePath(){
  return _serialize_path.getString();
}
void WSCconductor::setSerializePath(char* path){
  _serialize_path.setString(path);
}
WSDserialize* WSCconductor::beginTransaction(char* database,char* fname){
  WSDserialize* serialize = getSerialize(database);
  if (serialize == NULL){
WSMFtrace("WSCconductor::beginTransaction()  Data base does not exist.\n");
    return NULL;
  }
  serialize->setPath(getSerializePath());

  long ret = serialize->beginTransaction(fname);
  if (ret != WS_NO_ERR){
WSMFtrace("WSCconductor::beginTransaction()  Data base error.\n");
    return NULL;
  }
  return serialize;
}
long WSCconductor::endTransaction(WSDserialize* serialize){
  if (serialize->underTransaction() != False){
    serialize->endTransaction();
  }
  return WS_NO_ERR;
}
long WSCconductor::save(WSDserialize* serialize,char* type, char* name,void* ptr){
  return serialize->save(type,name,ptr);
}
long WSCconductor::load(WSDserialize* serialize, char* type, char* name,void* ptr){
  return serialize->load(type,name,ptr);
}
long WSCconductor::loadAlloc(WSDserialize* serialize, char* type, char* name,void** ptr,void* cdata){
  return serialize->loadAlloc(type,name,ptr,cdata);
}
long WSCconductor::saveGUI(WSDserialize* serialize, char* name,WSCbase* ptr){
  return serialize->save(ptr->getClassName(),name,(void*)ptr);
}
long WSCconductor::loadGUI(WSDserialize* serialize,WSCbase** ptr,WSCbase* cdata){
  return serialize->loadAlloc("WSCbase","*",(void**)ptr,(void*)cdata);
}
WSCbinType::WSCbinType(){
  long data = 0x11223344;
  char c[4];
  *(long*)c = data;
  if (c[0] == 0x44){
    _need_swap = True;
  }else{
    _need_swap = False;
  }
}
void WSCbinType::memcpy2b(char* ds,short sr){
  if (_need_swap == False){
    memcpy(ds,&sr,sizeof(short));
  }else{
    unsigned char* from;
    unsigned char to[2];
    from = (unsigned char*)&sr;
    ((unsigned char*)to)[1] = from[0];
    ((unsigned char*)to)[0] = from[1];
    memcpy(ds,to,sizeof(short));
  }
}
void WSCbinType::memscan2b(short* out,char* from){
  if (_need_swap == False){
    memcpy(out,from,sizeof(short));
  }else{
    char to[2];
    ((unsigned char*)to)[1] = from[0];
    ((unsigned char*)to)[0] = from[1];
    memcpy(out,to,sizeof(short));
  }
}
void WSCbinType::memscan4b(long* out,char* from){
  if (_need_swap == False){
    memcpy(out,from,sizeof(int));
  }else{
    unsigned char to[4];
    ((unsigned char*)to)[3] = from[0];
    ((unsigned char*)to)[2] = from[1];
    ((unsigned char*)to)[1] = from[2];
    ((unsigned char*)to)[0] = from[3];
    memcpy(out,to,sizeof(int));
  }
}
void WSCbinType::memscan8b(void* out,char* from){
  if (_need_swap == False){
    memcpy(out,from,sizeof(char)*8);
  }else{
    char to[8];
    long i;
    for(i=0; i<8; i++){
      ((unsigned char*)to)[7-i] = from[i];
    }
    memcpy(out,to,sizeof(char)*8);
  }
}
void WSCbinType::memscan16b(void* out,char* from){
  if (_need_swap == False){
    memcpy(out,from,sizeof(char)*16);
  }else{
    unsigned char to[16];
    long i;
    for(i=0; i<16; i++){
      ((unsigned char*)to)[15-i] = from[i];
    }
    memcpy(out,to,sizeof(char)*16);
  }
}
void WSCbinType::memscan32b(void* out,char* from){
  if (_need_swap == False){
    memcpy(out,from,sizeof(char)*32);
  }else{
    unsigned char  to[32];
    long i;
    for(i=0; i<32; i++){
      ((unsigned char*)to)[31-i] = from[i];
    }
    memcpy(out,to,sizeof(char)*32);
  }
}
void WSCbinType::memcpy4b(char* ds,long sr){
  if (_need_swap == False){
    memcpy(ds,&sr,sizeof(long));
  }else{
    unsigned char* from;
    unsigned char to[4];
    from = (unsigned char*)&sr;
    ((unsigned char*)to)[3] = from[0];
    ((unsigned char*)to)[2] = from[1];
    ((unsigned char*)to)[1] = from[2];
    ((unsigned char*)to)[0] = from[3];
    memcpy(ds,to,sizeof(long));
  }
}
void WSCbinType::memcpy8b(char* ds,void* sr){
  if (_need_swap == False){
    memcpy(ds,sr,sizeof(char)*8);
  }else{
    unsigned char *from,to[8];
    from = (unsigned char*)sr;
    ((unsigned char*)to)[7] = from[0];
    ((unsigned char*)to)[6] = from[1];
    ((unsigned char*)to)[5] = from[2];
    ((unsigned char*)to)[4] = from[3];
    ((unsigned char*)to)[3] = from[4];
    ((unsigned char*)to)[2] = from[5];
    ((unsigned char*)to)[1] = from[6];
    ((unsigned char*)to)[0] = from[7];
    memcpy(ds,to,sizeof(char)*8);
  }
}
void WSCbinType::memcpy16b(char* ds,void* sr){
  if (_need_swap == False){
    memcpy(ds,sr,sizeof(char)*16);
  }else{
    unsigned char *from,to[16];
    from = (unsigned char*)sr;
    long i;
    for(i=0; i<16; i++){
      ((unsigned char*)to)[15 -i] = from[i];
    }
    memcpy(ds,to,sizeof(char)*16);
  }
}
void WSCbinType::memcpy32b(char* ds,void* sr){
  if (_need_swap == False){
    memcpy(ds,sr,sizeof(char)*32);
  }else{
    unsigned char *from,to[32];
    from = (unsigned char*)sr;
    long i;
    for(i=0; i<32; i++){
      ((unsigned char*)to)[31 -i] = from[i];
    }
    memcpy(ds,to,sizeof(char)*32);
  }
}
