//
// Copyright (C) 1999-2002 Toshikaz Hirabayashi
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to
// deal in the Software without restriction, including without limitation the
// rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
// sell copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
// TOSHIKAZ HIRABAYASHI BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
// WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF
// OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.
//
// Except as contained in this notice, the name of Toshikaz Hirabayashi shall
// not be used in advertising or otherwise to promote the sale, use or other
// dealings in this Software without prior written authorization from
// Toshikaz Hirabayashi.

#include <x11/WSDxappDev.h>
#include <WScom.h>
#include <WSDfile.h>
#include <x11/WSDunixExecute.h>
#include <stdio.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/signal.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#include <WSCbaseList.h>
#include <WSDfileSystem.h>
#include <errno.h>

WSMFclassInit(WSDunixExecute,WSDexecute);

WSDexecute* _unix_execute_chandler(){
  return new WSDunixExecute;
}
class _unix_execute_init {
  public: _unix_execute_init(){
    WSDexecute::setCreateInstanceHandler((void*)_unix_execute_chandler);
  };
};
static _unix_execute_init  _unix_execute_init_execute;

WSDunixExecute::WSDunixExecute(){
  _xid = 0;
}
WSDunixExecute::~WSDunixExecute(){
  if (_pid != 0){
    killProcess();
  }
  if (_handle != NULL){
    pclose((FILE*)_handle);
    if (_mode != WS_EXEC_WR  && _xid != 0){
      XtRemoveInput(_xid);
      _xid = 0;
    }
    _handle = NULL;
    return;
  }
}
void WSDunixExecute::_callback(XtPointer ptr,int*,XtInputId*){
  WSDunixExecute* obj = (WSDunixExecute*)ptr;
  if (obj->_data_receive != NULL){
    obj->_data_receive(obj);
    WSGIappObjectList()->execUpdate();
  }
}
long WSDunixExecute::execute(char* exename,long mode){
  if (exename == NULL){
    return WS_ERR;
  }
//printf("WSDunixExecute::execute #%s#\n",exename);
  _exe_name = WSGIappFileSystem()->adjustFileName(exename);
  if (strcmp((char*)_dir_name,"")){
    int ret = chdir((char*)_dir_name);
    if (ret != 0){
      return WS_ERR;
    }
  }
//printf("WSDunixExecute::execute2 #%s#\n",(char*)_exe_name);
  if (mode == WS_EXEC_RDWR || mode == WS_EXEC_RD){
    if (mode == WS_EXEC_RD){
      _handle = popen(_exe_name,"r");
    }else if (mode == WS_EXEC_WR){
      _handle = popen(_exe_name,"w");
    }else{
      _handle = popen(_exe_name,"rw");
    }
    if (_handle == NULL){
      return WS_ERR;
    }
    if (mode != WS_EXEC_WR && _data_receive){
//      fcntl(fileno((FILE*)_handle),F_SETFL,O_NDELAY);
      _xid = XtAppAddInput(WSGIxwinAppDev()->appContext(),
                fileno((FILE*)_handle),
                (void*)(XtInputReadMask | XtInputWriteMask),
                 _callback,(void*)this);
     }
    _mode = mode;
    return WS_NO_ERR;
  }

  long num = _exe_name.getWords();
  if (num == 1){
#if 0
    if (file != NULL){
      struct stat sret;
      if (stat(file,&sret) == -1){
        return WS_ERR;
      }else if ((sret.st_mode & S_IFMT) == S_IFDIR){
        return WS_ERR;
      }else if (!(sret.st_mode & S_IFREG) || access(file,X_OK) != 0){
         return WS_ERR;
      }
    }else{
      return WS_ERR;
    }
#endif
    _pid  = (int)fork();
    if (_pid == 0){
      int ret = execl((char*)_exe_name,exename,NULL);
      if (ret == -1){
        exit(127);
      }else{
        exit(0);
      }
    }
  }else if (num > 1){
//printf("file=#%s#\n",(char*)_exe_name);
    char** args = new char*[num+1];
    long i;
    for(i=0; i<num; i++){
      args[i] = WSGFstrdup((char*)_exe_name.getWord(i));
//printf("args[%d]=%s\n",i,args[i]);
    }
    args[num] = NULL;
    _pid  = (int)fork();
    if (_pid == 0){
      int ret = execv(args[0],args);
      if (ret == -1){
        exit(127);
      }else{
        exit(0);
      }
    }
    for(i=0; i<num; i++){
      delete args[i];
    }
    delete args;
    int status;
    waitpid(_pid,&status,WNOHANG);
    if (WEXITSTATUS(status) == 127){
      return WS_ERR;
    }
  }
  return WS_NO_ERR;
}
long WSDunixExecute::killProcess(){
  if (_handle != NULL){
    if (_mode != WS_EXEC_WR && _xid != 0){
      XtRemoveInput(_xid);
      _xid = 0;
    }
    pclose((FILE*)_handle);
    _handle = NULL;
    return WS_NO_ERR;
  }
  if (_pid == 0){
    return WS_NO_ERR;
  }
  kill(_pid,SIGKILL);
  while(1){
    int dust;
    if ( (int)waitpid(_pid,&dust,0) < 0){
       continue;
    }
    break;
  }
  _pid = 0;
  return WS_NO_ERR;
}

long WSDunixExecute::read(char* buf,long len){
  if (_mode == WS_EXEC_RD || _mode == WS_EXEC_RDWR){
    if (_handle != NULL){
      int fdno = fileno((FILE*)_handle);
      int ret = ::read(fdno,buf,len);
      return ret;
    }
  }
  return -1;
}
char* WSDunixExecute::gets(char* buf,long len){
  if (_mode == WS_EXEC_RD || _mode == WS_EXEC_RDWR){
    if (_handle != NULL){
      return fgets(buf,len,(FILE*)_handle);
    }
  }
  return 0;
}


long WSDunixExecute::write(char* buf,long len){
  if (_mode == WS_EXEC_WR || _mode == WS_EXEC_RDWR){
    if (_handle != NULL){
      int fdno = fileno((FILE*)_handle);
      return ::write(fdno,buf,len);
    }
  }
  return -1;
}


